#include <iostream>

using namespace std;

// This struct is used to store the optimal solution to sub problems
struct cuts {
    // Initialize the cost to negative infinity (INT_MIN) and the firstCut and laterCuts to 0 (no cut)
    int cost = INT_MIN;
    int firstCut = 0;
    int laterCuts = 0;
};

// lengthOfRod is the number of possible segments, or the number of possible cuts plus 1
// p is an array containing costs. Note that p must include a cost for a 0 length segment, which is 0
void optimalCuts(int lengthOfRod, int* p) {
    // Create an array to store results
    cuts* r = new cuts[lengthOfRod + 1];
    // Set the cost of a 0 length rod to 0
    r[0].cost = 0;
    // Iterate through all starting rod sizes
    for (int i = 1; i <= lengthOfRod; i++) {
        // Iterate through every later cut available
        for (int j = 1; j <= i; j++) {
            // If it is a better solution, set the cost, first cut, and later cuts
            if (r[i].cost < p[j] + r[i - j].cost) {
                r[i].cost = p[j] + r[i - j].cost;
                r[i].firstCut = j;
                r[i].laterCuts = i - j;
            }
        }
    }
    // Print the results
    cout << "Cost: " << r[lengthOfRod].cost << endl;
    // To print the required results, we print firstCut, then recurse into laterCuts until we reach a first cut of 0 (no cut)
    cout << "Cuts: ";
    cuts c = r[lengthOfRod];
    while (c.firstCut != 0) {
        cout << c.firstCut << " ";
        c = r[c.laterCuts];
    }
    cout << endl;
    // Don't forget to clean up!
    delete[] r;
}

int main()
{
    int lengthOfRod1 = 8;
    int p1[] = {0, 2, 5, 8, 9, 10, 16, 17, 18};
    optimalCuts(lengthOfRod1, p1);
    cout << endl;

    int lengthOfRod2 = 16;
    int p2[] = { 0, 2, 5, 8, 9, 10, 21, 17, 20, 24, 30, 35, 38, 40, 42, 45, 40 };
    optimalCuts(lengthOfRod2, p2);
    cout << endl;

    int lengthOfRod3 = 10;
    int p3[] = { 0, 1, 5, 8, 9, 10, 17, 17, 20, 24, 30 };
    optimalCuts(lengthOfRod3, p3);
    cout << endl;
}

int max(int num1, int num2) {
    return num1 > num2 ? num1 : num2;
}